// mds.h
// almost imported from MDSviewer source code
// Definition of MDS Structure


struct MDSTag
{
	char		cName[64];    // Name of tag
	float		fTorsoWeight; // 0.0, 0.3
	long		lBoneIndex;   // Index of bone this tag is attached to
}; 

struct MDSBone
{
	char		cName[64];      // Bone name
	long		lParentIndex;   // Bone parent index (-1 if root)
	float		fTorsoWeight;   // 0.0 to 1.0
	float		fParentDist;    // Distance from parent bone to this bone's pivot point
	long		ulFlags;        // Bit 0 is set if bone is a tag
  };
 
struct Coord
{
	float		x;
	float		y;
	float		z;
};

struct MDSCompressedBoneFrame
{
	short		sPitch;        // Angle
	short		sYaw;          // Angle
	short		sRoll;         // Angle
	short		sZero;         // Unused - normally zero (for DWORD alignment)
	short		sOffsetPitch;  // 
	short		sOffsetYaw;
};

struct MDSFrame
{
	Coord		bboxMin;      // Bounding box min
	Coord		bboxMax;      // Bounding box max
	Coord		LocalOrigin;  // Local origin
	float		fRadius;      // Radius of bounding sphere (used for runtime culling)
	Coord		ParentOffset; // Offset of parent bone from origin

    // Following this structure is a list of compressed bone frames
    //CompressedBoneFrame  bones[1];
};

struct TexCoord
{
	float		u;
	float		v;
};

struct MDSWeight
{
	long		lBoneIndex;  // Index of bone in the main bone list
	float		fWeight;     // Bone weighting (from 0.0 to 1.0)
	Coord		xyz;         // xyz bone space position of vertex
};

struct MDSVertex
{
	Coord		Normal;       // Vertex normal vector. In base pose/model space I presume.
	TexCoord	TexCoords;    // Texture (s,t) co-ordinates
	long		lNumWeights;  // Number of bone weights affecting this vertex

	long		lFixedParent; // Stay equidistant from this parent
	float		fFixedDist;   // Fixed distance from parent
	//Weight	MDSWeights[1];   // List of weights follows
};

struct MDSTriangle
{
	unsigned long	ulVertex1;
	unsigned long	ulVertex2;
	unsigned long	ulVertex3;
};

struct MDSSurface
{
	long		lId;             // Always 0x0008
	char		cName[64];       // Name of surface
	char		cShader[64];     // Name of shader
	long		lShaderIndex;    // Used in game only

	long		lMinLod;         // Minimum Lod level
	long		lOffsetHeader;   // Offset of this surface header, always negative

	long		lNumVertices;    // Number of vertices

	long		lOffsetVertices; // Offset of vertices (MDSVertex * numVerts)

	long		lNumTriangles;   // Number of triangles

	long		lOffsetTriangles;// Offset of triangles (MDSTriangle * numTris)

	long		lOffsetCollapseMap; // Offset of the collapse map (uint32 * numVerts in size)

	long		lNumBoneRefs;    // Number of bone references (bones that influence this surface)

	long		lOffsetBoneRefs; // Offset of bone refs (uint32 * numBones)

	long		lOffsetEnd;      // Offset of the end of this surface, from start of MDSSurface header.

};

struct MDSHeader
{
	long		lId;              // "MDSW"
	long		lVersion;        // 0x0004
	char		cName[64];       // Model name eg. "body.mds"

	float		fLodScale;       // LOD Scale
	float		fLodBias;        // LOD Bias

	long		lNumFrames;      // Number of animation frames
	long		lNumBones;       // Number of bones

	long		lOffsetFrames;   // Offset of animation frames (see MDSFrame)

	long		lOffsetBones;    // Offset of bones (see MDSBone)

	long		lTorsoParent;    // Index of torso parent bone

	long		lNumSurfaces;    // Number of surfaces

	long		lOffsetSurfaces; // Offset of surfaces (numSurfaces * MDS_Surface)

	long		lNumTags;        // Number of tags

	long		lOffsetTags;     // Offset of tags (numTags * MDS_Tag)

	long		lOffsetEnd;      // Offset of end of file

};

// EOF