
/*
main.cpp

Source code for MDSMX tool

2004/03/16 ikanatto

No copyrights
*/

/*
########################################################

Headers and Definitions

########################################################
*/

#include "stdafx.h"
#include "stdio.h"
#include "io.h"
#include "math.h"
#include "fcntl.h"
#include "sys\stat.h"
#include "mdm.h"
#include "mdx.h"
#include "mds.h"

#define		MAXMDSSIZE		10000000	// 10MB
#define		MAXMDMSIZE		1000000		// 1MB
#define		MAXMDXSIZE		10000000	// 10MB
#define		MAXBONE			128		// Default MDS limits
#define		MAXFRAME		10000		// or more?
#define		MDSVERSION		4
#define		MDMVERSION		3
#define		MDXVERSION		2

#define		TRUE			1
#define		FALSE			0

struct Flags
{

	int		MODE;
	char	MDSFILENAME[256];
	char	MDMFILENAME[256];
	char	MDXFILENAME[256];
	int		REPORT;
	int		BASEFRAME;
	int		REPLACEMDMNAME;
	char	MDMNAME[64];
	int		REPLACEMDXNAME;
	char	MDXNAME[64];
	int		RESCALEFACTOR;
	char	REFERENCEMDM[64];
	char	REFERENCEMDX[64];
	char	MDMSOURCE[64];
	char	MDXSOURCE[64];
	int		LOCALORIGINZ;
    
};

char Parse_MDSBone( MDSBone * , int , long * , long * );

int Load_MDS( char * , char * );
int Load_MDM( char * , char * );
int Load_MDX( char * , char * );
int Report_MDM( char * );
int Convert_MDS_to_MDX( char * , char * );
int Convert_MDS_to_MDM( char * , char * , int );
int Replace_Tag( char * , char * ); 
int Rescale_MDM( char * , int );
int Replace_Tag_All( char * , char * );
int Replace_Bone( char * , char * );
int Rescale_MDX( char * , int );
int Save_File( char * , char * , int );
int strequ( char * , char * );
int Parse_Configuration( char * , Flags * );
int a2i( char * );

void yaw( Point3 * , short int );
void pitch( Point3 * , short int );
void roll( Point3 * , short int );
void prt( char * );
void Print_Usage();

/*
########################################################

main

Strats here :p

########################################################
*/

int _tmain( int argc,  char *argv[] )
{

	printf("\n MDS convertion tool ver. 1.00\n\n");

	/* Initialzing */

	char	*mds		=	(char *) malloc( MAXMDSSIZE );
	char	*mdx		=	(char *) malloc( MAXMDXSIZE ); // should be changed for smaller
	char	*mdm		=	(char *) malloc( MAXMDMSIZE ); // 
	char	*refmdx		=	(char *) malloc( MAXMDXSIZE );
	char	*refmdm		=	(char *) malloc( MAXMDMSIZE );

	char 	*mdsfilename 	= 	"body.mds"; // these are default filenames to save each model data
	char 	*mdmfilename 	= 	"body.mdm";
	char 	*mdxfilename 	= 	"body.mdx";

	Flags	flags;
	char	*cfgfilename	=	"mdsmx.cfg";

	int mdssize, mdxsize, mdmsize;

	/* Parse Config */

	if( Parse_Configuration( cfgfilename , &flags ) == -1 )
	{

		printf("Error found in config file. Check %s.\n", cfgfilename);
		exit(0);

	}

	/* Switch */

	if( ( flags.MODE == 1 ) | ( flags.MODE == 3 ) ) // MDS->MDM
	{

		if( ( mdssize = Load_MDS( flags.MDSFILENAME , mds ) ) == -1 )
		{
			printf("ERROR:Failed to load MDS.\n");
			exit(0);
		}

		if( ( mdmsize = Convert_MDS_to_MDM( mds , mdm , flags.BASEFRAME) ) == -1 )
		{
			printf("ERROR:Failed to Convert MDS to MDM.\n");
			exit(0);
		}

		if( Save_File( mdm , flags.MDMFILENAME , mdmsize ) == -1)
		{
			printf("ERROR:Failed to Save File.\n");
			exit(0);
		}

	}

	if( (flags.MODE == 2) | (flags.MODE == 3) ) //MDS->MDX
	{

		if( ( mdssize = Load_MDS( flags.MDSFILENAME , mds ) ) == -1 )
		{
			printf("ERROR:Failed to load MDS.\n");
			exit(0);
		}

		if( ( mdxsize = Convert_MDS_to_MDX( mds , mdx ) ) == -1 )
		{
			printf("ERROR:Failed to Convert MDS to MDM.\n");
			exit(0);
		}

		if( Save_File( mdx , flags.MDXFILENAME , mdxsize ) == -1)
		{
			printf("ERROR:Failed to Save File.\n");
			exit(0);
		}

	}

	if( ( (flags.MODE == 1) | (flags.MODE == 3) ) && ( flags.REPORT == 1 ) ) Report_MDM( mdm );

	if( flags.MODE == 6 )
	{

		if( ( mdmsize = Load_MDM( flags.MDMSOURCE , mdm ) ) == -1 )
		{
			printf("ERROR:Failed to load MDM.\n");
			exit(0);
		}

		if( Rescale_MDM( mdm , flags.RESCALEFACTOR ) == -1 )
		{
			printf("ERROR: Failed to rescale MDM.\n");
			exit(0);
		}

		if( Save_File( mdm , flags.MDMFILENAME , mdmsize ) == -1 )
		{
			printf("ERROR:Failed to Save File.\n");
			exit(0);
		}

	}

	if( flags.MODE == 4 )
	{

		if( ( mdmsize = Load_MDM( flags.MDMSOURCE , mdm ) ) == -1 )
		{
			printf("ERROR:Failed to load MDM.\n");
			exit(0);
		}

		if( Load_MDM( flags.REFERENCEMDM , refmdm ) == -1 )
		{
			printf("ERROR:Failed to load Reference MDM.\n");
			exit(0);
		}

		if( Replace_Tag_All( mdm , refmdm ) == -1 )
		{
			printf("ERROR: Failed to replace tags.\n");
			exit(0);
		}

		MDMHeader	*mdmheader	= (MDMHeader *)mdm;
		mdmsize = mdmheader->OffsetEOF;

		if( Save_File( mdm , flags.MDMFILENAME , mdmsize ) == -1 )
		{
			printf("ERROR: Failed to save file.\n");
			exit(0);
		}
		
	}
    
	if( flags.MODE == 5 )
	{

		if( ( mdxsize = Load_MDX( flags.MDXSOURCE , mdx ) ) == -1 )
		{
			printf("ERROR:Failed to load MDX.\n");
			exit(0);
		}

		if( Load_MDX( flags.REFERENCEMDX , refmdx ) == -1 )
		{
			printf("ERROR:Failed to load Reference MDX.\n");
			exit(0);
		}

		if( Replace_Bone( mdx , refmdx ) == -1 )
		{
			printf("ERROR: Failed to replace bones.\n");
			exit(0);
		}

		if( Save_File( mdx , flags.MDXFILENAME , mdxsize ) == -1 )
		{
			printf("ERROR: Failed to save file.\n");
			exit(0);
		}
		
	}

	if( flags.MODE == 7 )
	{

		if( ( mdxsize = Load_MDX( flags.MDXSOURCE , mdx ) ) == -1 )
		{
			printf("ERROR:Failed to load MDX.\n");
			exit(0);
		}

		if( Rescale_MDX( mdx , flags.RESCALEFACTOR ) == -1 )
		{
			printf("ERROR: Failed to rescale MDX.\n");
			exit(0);
		}

		if( Save_File( mdx , flags.MDXFILENAME , mdxsize ) == -1 )
		{
			printf("ERROR: Failed to save file.\n");
			exit(0);
		}
		
	}

	printf("\n\n --- All done ---\n\n");

	free( mdm );
	free( mds );
	free( mdx );

	return 1;

}

/*
########################################################

Print_Usage

Display available options

########################################################
*/

void Print_Usage()
{
	printf(" Usage: mdsmx");

	printf("\n Please visit http://mak.s13.xrea.com/ET/ for more info and bug report!\n\n");

}

/*
########################################################

Parse_Configuration

Loads cfg file and read 'key and value'. Stores them into flags. 
Returns -1 if error.

########################################################
*/

int Parse_Configuration( char *configfile , Flags *flags )
{

	FILE *fp;
	int i, j, argc, lines, err;
	char argv[1024];
	char key[256];
	char value[768];

	fp = fopen( configfile , "r" );

	if( fp == NULL )
	{
		printf("Failed to open config file.\n");
		return -1;
	}

	for( lines = 1; fgets( argv , 1024 , fp ) != NULL; lines++ ){

		for( i = 0; ( *(argv + i) != 0 ) && ( *(argv + i) != '\n' ); i++ );
        
		*(argv + i) = 0;

		//printf("argv(%2d):%s\n", lines , argv);

		err = 0;

		if( strlen( argv ) < 3 ) continue;

		if( (argv[0] == '/') && (argv[1] == '/') ) continue;

		for( i = 0; argv[i] != '='; i++ ){

			if( i > 254 ){

				printf("Parsing config error at line %d:Too long key entry.\n",lines);
				return -1;

			}

			key[i] = argv[i];

			if( i > strlen( argv ) - 2 ){

				printf("Parsing config error at line %d:No Value\n",lines);
				err = 1;
				return -1;

			}

		}

		key[i++] = 0; // skip '='

		for( j = 0; argv[i] != 0; j++ ) value[j] = argv[i++];

		value[j] = 0;

		/* Switch by key entry */

		//printf("key:%s value:%s\n",key,value);

		if( strequ( key , "MODE" ) )
		{
			
			flags->MODE = a2i( value );
			if( (flags->MODE < 1) | (flags->MODE > 7) )
			{
				printf("Parsing config error at line %d:Invalid MODE value\n",lines);
				return -1;
			}

		}

		else if( strequ( key , "MDS" ) )
		{
			
			for( i = 0; ( *(value + i) ) && ( i < 255 ); i++) flags->MDSFILENAME[i] = value[i];
			flags->MDSFILENAME[i] = 0;
		}

		else if( strequ( key , "MDM" ) )
		{
			
			for( i = 0; ( *(value + i) ) && ( i < 256 ); i++) flags->MDMFILENAME[i] = value[i];
			flags->MDMFILENAME[i] = 0;

		}

		else if( strequ( key , "MDX" ) )
		{

			for( i = 0; ( *(value + i) ) && ( i < 256 ); i++) flags->MDXFILENAME[i] = value[i];
			flags->MDXFILENAME[i] = 0;

		}

		else if( strequ( key , "REPORT" ) )
		{

			flags->REPORT = a2i( value );

		}

		else if( strequ( key , "BASEFRAME" ) )
		{

			flags->BASEFRAME = a2i( value );

		}

		else if( strequ( key , "REPLACEMDMNAME" ) )
		{

			flags->REPLACEMDMNAME = a2i( value );

		}

		else if( strequ( key , "MDMNAME" ) )
		{

			for( i = 0; ( *(value + i) ) && ( i < 64 ); i++) flags->MDMNAME[i] = value[i];
			flags->MDMNAME[i] = 0;

		}

		else if( strequ( key , "REPLACEMDXNAME" ) )
		{

			flags->REPLACEMDXNAME = a2i( value );

		}

		else if( strequ( key , "MDXNAME" ) )
		{

			for( i = 0; ( *(value + i) ) && ( i < 64 ); i++) flags->MDXNAME[i] = value[i];
			flags->MDXNAME[i] = 0;

		}

		else if( strequ( key , "REFERENCEMDM" ) )
		{

			for( i = 0; ( *(value + i) ) && ( i < 63 ); i++) flags->REFERENCEMDM[i] = value[i];
			flags->REFERENCEMDM[i] = 0;

		}

		else if( strequ( key , "REFERENCEMDX" ) )
		{

			for( i = 0; ( *(value + i) ) && ( i < 63 ); i++) flags->REFERENCEMDX[i] = value[i];
			flags->REFERENCEMDX[i] = 0;

		}

		else if( strequ( key , "MDMSOURCE" ) )
		{

			for( i = 0; ( *(value + i) ) && ( i < 63 ); i++) flags->MDMSOURCE[i] = value[i];
			flags->MDMSOURCE[i] = 0;

		}

		else if( strequ( key , "MDXSOURCE" ) )
		{

			for( i = 0; ( *(value + i) ) && ( i < 63 ); i++) flags->MDXSOURCE[i] = value[i];
			flags->MDXSOURCE[i] = 0;

		}

		else if( strequ( key , "RESCALEFACTOR" ) )
		{
			
			flags->RESCALEFACTOR = a2i( value );

		}

		else
		{
			printf("Parsing config error at line %d:No such key(%s).\n",lines,key);
			return -1;

		}

	}

	fclose( fp );

	return 1;

}

/*
########################################################

General Functions

########################################################
*/
void prt( char * )
{
	return;
}

int strequ( char *str1 , char *str2 )
{

	while( *str1 )	if( *str1++ != *str2++ ) return 0;

	if( *str2 != 0 ) return 0;

	return 1;

}

int a2i( char *str )
{

	int i = 0;

	if( ( *str < '0' ) | ( *str > '9' ) ) return -1;

	while( ( *str >= '0' ) && ( *str <= '9' ) ) i = 10 * i + ( *str++ - '0' );

	return i;

}




/*
########################################################

Convert_MDS_to_MDX

Converts MDS to MDX and returns MDX size or -1 if error.

########################################################
*/

int Convert_MDS_to_MDX( char *mdsbuf , char *mdxbuf )
{

	printf("\n--- Build MDX ---\n\n");

	/* Set address to pointers */

	MDSHeader	*mdsheader		=	(MDSHeader *)((char *)mdsbuf );
	MDSFrame	*mdsframe		=	(MDSFrame *)((char *)mdsbuf + mdsheader->lOffsetFrames );
	MDSBone		*mdsbone		=	(MDSBone *)((char *)mdsbuf + mdsheader->lOffsetBones );
	MDSTag		*mdstag			=	(MDSTag *)((char *)mdsbuf + mdsheader->lOffsetTags );
	MDSSurface	*mdssurface		=	(MDSSurface *)((char *)mdsbuf + mdsheader->lOffsetSurfaces );

	/* Init general variables */

	int		i, j = 0, k, l, m;

	char		mdxbonecount = 0;
	long		mdmbonetomds[MAXBONE]; // MDS Bone number -> MDM
	long		mdsbonetomdm[MAXBONE]; // MDM Bone number -> MDS

	/* Parse MDSBone */

	mdxbonecount = Parse_MDSBone( mdsbone , mdsheader->lNumBones , mdmbonetomds , mdsbonetomdm );

	MDXHeader	*mdxheader	=	(MDXHeader *) ((char *) mdxbuf );
	MDXFrame	*mdxframe	=	(MDXFrame *) ((char *) mdxbuf + sizeof(MDXHeader) );
	MDXBone		*mdxbone	=	(MDXBone *) ((char *) mdxbuf + sizeof(MDXHeader) + mdsheader->lNumFrames * ( sizeof(MDXFrame) + mdxbonecount * sizeof(MDXCompFrame) ) );

	/* Offset to each frame set */

	long		OffsetNextFrameX	=	sizeof(MDXFrame) + mdxbonecount * sizeof(MDXCompFrame);
	long		OffsetNextFrameS	=	sizeof(MDSFrame) + mdsheader->lNumBones * sizeof(MDSCompressedBoneFrame);

	/* Construct MDX */

	printf("Header...");

	/* Header */

	char		*mdxident	=	"MDXW";
	i = 0;
	while( *mdxident ) mdxheader->MDXIdent[i++] = *mdxident++;

	mdxheader->MDXVersion		=	2;

	char		*mdxname	=	"output/animations/human/base/body.mdx";
	i = 0;
	while( *mdxname ) mdxheader->MDXName[i++] = *mdxname++;
	while(  i < 64  ) mdxheader->MDXName[i++] = '\0';

	mdxheader->NumFrame		=	mdsheader->lNumFrames;

	mdxheader->NumBone		=	mdxbonecount;

	mdxheader->OffsetFrame		=	sizeof(MDXHeader);

	mdxheader->OffsetBone		=	sizeof(MDXHeader) + mdxheader->NumFrame * ( sizeof(MDXFrame) + mdxbonecount * sizeof(MDXCompFrame) );

	mdxheader->TorsoParent		=	mdsheader->lTorsoParent;

	mdxheader->OffsetEOF		=	sizeof(MDXHeader) + mdxheader->NumFrame * ( sizeof(MDXFrame) + mdxbonecount * sizeof(MDXCompFrame) ) + mdxbonecount * sizeof(MDXBone);

	printf("Done\n");
	/* 
	Frame 

	There is no difference between the structure of MDS Frame and that of MDX.
	Besides that the scaling or something, just copying them.

	Compressed Bone Frame

	Because of lack of the tags as bone, we need to pass the data of tag CBF.
	*/

	printf("Frames...");

	MDXFrame	*currentmdxframe	=	(MDXFrame *) mdxframe;
	MDSFrame	*currentmdsframe	=	(MDSFrame *) mdsframe;

	for( i = 0; i < mdxheader->NumFrame; i++ )
	{

	currentmdxframe = (MDXFrame *)((char *)mdxframe + OffsetNextFrameX * i);
	currentmdsframe = (MDSFrame *)((char *)mdsframe + OffsetNextFrameS * i);

	*currentmdxframe = *((MDXFrame *)currentmdsframe);

		for( j = 0; j < mdxheader->NumBone; j++ )
		{

			*((MDXCompFrame *)((char *)currentmdxframe + sizeof(MDXFrame) + sizeof(MDXCompFrame) * j)) = *((MDXCompFrame *)((char *)currentmdsframe + sizeof(MDSFrame) + sizeof(MDSCompressedBoneFrame) * *(mdmbonetomds + j)));

		}

	}

	printf("Done\n");

	printf("Bones...");

	MDXBone		*currentmdxbone		=	(MDXBone *)mdxbone;
	MDSBone		*currentmdsbone		=	(MDSBone *)mdsbone;

	for( i = 0; i < mdxheader->NumBone; i++)
	{

		currentmdsbone		=	(MDSBone *)((char *)mdsbone + sizeof(MDSBone) * *(mdmbonetomds + i));
		currentmdxbone		=	(MDXBone *)((char *)mdxbone + sizeof(MDXBone) * i);

		*currentmdxbone = *((MDXBone *)currentmdsbone);

	/* Exchange MDX Bone Parent Index from MDS Index*/

		if( currentmdxbone->ParentIndex != -1 ){

			currentmdxbone->ParentIndex = (long) *(mdsbonetomdm + currentmdxbone->ParentIndex);

		}

	}

	printf("Done\n");

	printf("MDX Convertion has been done.\n");

	printf("MDX size			:%d\n",mdxheader->OffsetEOF );

	return mdxheader->OffsetEOF;

}

/*
########################################################

Convert_MDS_to_MDM

Converts MDS to MDM and returns MDM size. If error, returns -1.

########################################################
*/

int Convert_MDS_to_MDM( char *mdsbuf, char *mdmbuf , int baseframenumber )
{

	printf("\n--- Build MDM ---\n\n");

	/* Set address to pointers */

	MDSHeader	*mdsheader		=	(MDSHeader *)((char *)mdsbuf );
	MDSFrame	*mdsframe		=	(MDSFrame *)((char *)mdsbuf + mdsheader->lOffsetFrames );
	MDSBone		*mdsbone		=	(MDSBone *)((char *)mdsbuf + mdsheader->lOffsetBones );
	MDSTag		*mdstag			=	(MDSTag *)((char *)mdsbuf + mdsheader->lOffsetTags );
	MDSSurface	*mdssurface		=	(MDSSurface *)((char *)mdsbuf + mdsheader->lOffsetSurfaces );

	MDMHeader 	*mdmheader 		= 	(MDMHeader *)mdmbuf;
	MDMSurfaceHeader *mdmsurface 		= 	(MDMSurfaceHeader *)( (char *)mdmbuf + sizeof(MDMHeader) );


	/* Init general variables */

	int		i, j = 0, k, l, m;

	char		mdxbonecount = 0;
	long		mdmbonetomds[MAXBONE]; // MDS Bone number -> MDM
	long		mdsbonetomdm[MAXBONE]; // MDM Bone number -> MDS

	int		totalvertex = 0;
	int		currentsurfacesize = 0, totalsurfacesize = 0;

	/* Parse MDSBone */

	mdxbonecount = Parse_MDSBone( mdsbone , mdsheader->lNumBones , mdmbonetomds , mdsbonetomdm );

	/*
	Surface

	There are a few differences between MDS and MDM in Surface.
	Vertex does not have Fixed Distance property in MDM, so removing them will go woot.
	Also we change the MDS bone number to MDM's.
	*/

	printf("Surfaces...");
	
	/* Handling pointers */

	MDSSurface			*currentmdssurface;
	MDMSurfaceHeader		*currentmdmsurface;
	MDSVertex			*currentmdsvertex;
	MDMVertex			*currentmdmvertex;
	MDSWeight			*currentmdsweight;
	MDMWeight			*currentmdmweight;
	MDSTriangle			*currentmdstriangle;
	MDMTriangle			*currentmdmtriangle;
	char				*currentmdscollapsemap;
	char				*currentmdmcollapsemap;
	long				*currentmdsboneref;
	long				*currentmdmboneref;

	/* Convert Surface */

	currentmdssurface = mdssurface;
	currentmdmsurface = mdmsurface;

	for( i = 0; i < mdsheader->lNumSurfaces; i++)
	{

		/* Surface Header */

		currentmdmsurface->SurfaceIdent = 9;
		
		for( l = 0; l < 64; l++ ) currentmdmsurface->SurfaceName[l] = currentmdssurface->cName[l];

		for( l = 0; l < 64; l++ ) currentmdmsurface->ShaderName[l] = currentmdssurface->cShader[l];

		currentmdmsurface->ShaderIndex = 0/*currentmdssurface->lShaderIndex*/;

		currentmdmsurface->LoDMin = currentmdssurface->lMinLod;

		currentmdmsurface->OffsetSurfaceHeader = -(sizeof(MDMHeader) + totalsurfacesize);

		currentmdmsurface->NumVertex = currentmdssurface->lNumVertices;

		currentmdmsurface->OffsetVertex = sizeof(MDMSurfaceHeader);

		currentmdmsurface->NumTriangle = currentmdssurface->lNumTriangles;

		currentmdmsurface->OffsetTriangle = currentmdmsurface->OffsetVertex + ( currentmdssurface->lOffsetTriangles - currentmdssurface->lOffsetVertices ) - 2 * 4 * currentmdssurface->lNumVertices;

		currentmdmsurface->OffsetCollapsemap = currentmdmsurface->OffsetTriangle + ( currentmdssurface->lOffsetCollapseMap - currentmdssurface->lOffsetTriangles);

		currentmdmsurface->NumBoneref = currentmdssurface->lNumBoneRefs;

		currentmdmsurface->OffsetBoneref = currentmdmsurface->OffsetCollapsemap + (currentmdssurface->lOffsetBoneRefs - currentmdssurface->lOffsetCollapseMap);

		currentmdmsurface->OffsetNextSurface = currentmdmsurface->OffsetBoneref + (currentmdssurface->lOffsetEnd - currentmdssurface->lOffsetBoneRefs);

		currentsurfacesize += sizeof(MDMSurfaceHeader);

		//printf("SURFACE #%d:%s\n", i , currentmdmsurface->SurfaceName );

		//printf("[SURFACE VERTEX] (SIZE:%d) (OFFSET:%d)\n",currentsurfacesize, currentmdmsurface->OffsetVertex );

		/* Vertex */

		currentmdmvertex = (MDMVertex *) ((char *)currentmdmsurface + sizeof(MDMSurfaceHeader));
		currentmdsvertex = (MDSVertex *) ((char *)currentmdssurface + sizeof(MDSSurface));
		
		for( j = 0; j < currentmdssurface->lNumVertices; j++ )
		{

			currentmdmvertex->Normal.x = currentmdsvertex->Normal.x;
			currentmdmvertex->Normal.y = currentmdsvertex->Normal.y;
			currentmdmvertex->Normal.z = currentmdsvertex->Normal.z;

			currentmdmvertex->TextureCoord.u = currentmdsvertex->TexCoords.u;
			currentmdmvertex->TextureCoord.v = currentmdsvertex->TexCoords.v;

			/* MDM does not have FixedParentDistance and FixedParentBoneIndex. This means that a vertex might go far from its estimated place. */

			currentmdmvertex->NumBoneWeight = currentmdsvertex->lNumWeights;

			/* Weight */

			currentmdsweight = (MDSWeight *)((char *)currentmdsvertex + sizeof(MDSVertex));
			currentmdmweight = (MDMWeight *)((char *)currentmdmvertex + sizeof(MDMVertex));

			for( k = 0; k < currentmdmvertex->NumBoneWeight; k++)
			{

				currentmdmweight->BoneIndex = *(mdsbonetomdm + currentmdsweight->lBoneIndex); // eek
				
				currentmdmweight->Weight = currentmdsweight->fWeight;

				currentmdmweight->Coord.x = currentmdsweight->xyz.x;
				currentmdmweight->Coord.y = currentmdsweight->xyz.y;
				currentmdmweight->Coord.z = currentmdsweight->xyz.z;

				/* Next Weight */

				currentmdmweight++;
				currentmdsweight++;

			}

			currentsurfacesize += currentmdmvertex->NumBoneWeight * sizeof(MDMWeight);

			/* Next Vertex */
			currentmdmvertex = (MDMVertex *)currentmdmweight;
			currentmdsvertex = (MDSVertex *)currentmdsweight;

		}

		currentsurfacesize += currentmdmsurface->NumVertex * sizeof(MDMVertex);

		totalvertex += currentmdmsurface->NumVertex;

		//printf("[SURFACE TRIANGLE] (SIZE:%d) (OFFSET:%d)\n",currentsurfacesize, currentmdmsurface->OffsetTriangle );

		/* Triangle */

		currentmdstriangle = (MDSTriangle *)currentmdsvertex;
		currentmdmtriangle = (MDMTriangle *)currentmdmvertex;

		for( j = 0; j < currentmdssurface->lNumTriangles; j++ )
		{

			currentmdmtriangle->VertexIndex1 = currentmdstriangle->ulVertex1;
			currentmdmtriangle->VertexIndex2 = currentmdstriangle->ulVertex2;
			currentmdmtriangle->VertexIndex3 = currentmdstriangle->ulVertex3;

			/* Next Triangle */

			currentmdmtriangle++;
			currentmdstriangle++;

		}
		
		currentsurfacesize += currentmdssurface->lNumTriangles * sizeof(MDMTriangle);

		//printf("[SURFACE Collapsemap] (SIZE:%d) (OFFSET:%d)\n",currentsurfacesize, currentmdmsurface->OffsetCollapsemap );

		/* Collapsemap */

		currentmdscollapsemap = (char *)currentmdstriangle;
		currentmdmcollapsemap = (char *)currentmdmtriangle;

		for( j = 0; j < (currentmdssurface->lOffsetBoneRefs - currentmdssurface->lOffsetCollapseMap); j++)
		{

			*(currentmdmcollapsemap++) = *(currentmdscollapsemap++);
			
			currentsurfacesize += sizeof(char);

		}

		//printf("[SURFACE BONEREF] (SIZE:%d) (OFFSET:%d)\n",currentsurfacesize, currentmdmsurface->OffsetBoneref );

		/* Bone Reference */

		currentmdsboneref = (long *)currentmdscollapsemap;
		currentmdmboneref = (long *)currentmdmcollapsemap;

		for( j = 0; j < currentmdssurface->lNumBoneRefs; j++)
		{

			*(currentmdmboneref++) = *(mdsbonetomdm + *(currentmdsboneref++));
			
		}

		currentsurfacesize += currentmdssurface->lNumBoneRefs * sizeof(long);

		/* Next Surface */
		
		if( (((char *)currentmdmboneref - (char *)currentmdmsurface) != currentsurfacesize) | (currentmdmsurface->OffsetNextSurface != currentsurfacesize) )
		{

			printf("Oops, Offset does not match total size in #%d surface.(SIZE:%d)(OFFSET:%d)(ADDRESS:%d)\n",i,currentsurfacesize,currentmdmsurface->OffsetNextSurface,(char *)currentmdmboneref - (char *)currentmdmsurface);

		}
		
		currentmdssurface = (MDSSurface *)currentmdsboneref;
		currentmdmsurface = (MDMSurfaceHeader *)currentmdmboneref;

		totalsurfacesize += currentsurfacesize;
		currentsurfacesize = 0;

	}

	printf("Done.\n");

	/*
	MDM Tag

	MDM tag is completely different from MDS one but similar to MD3's.
	MDM tags is 'fixed'. Other words, they are no longer bones but such as surface vertices.
	Thus we need to do the followings.
	(1) make the "Bone Tree" that is junction of bones from a root bone to a tag
	(2) calcurate the normalized coordinates and origin position from a MDS Frame
	*/

	printf("Tags...");

	MDMTag					*mdmtag = (MDMTag *)( (char *)mdmbuf + sizeof(MDMHeader) + totalsurfacesize ); 

	/* Handling pointers */

	MDSTag			*currentmdstag = mdstag;
	MDMTag			*currentmdmtag = mdmtag;
	MDSBone			*currentmdsbone = mdsbone;
	MDSBone			*tempmdsbone;
	MDSFrame		*tempmdsframe = mdsframe;
	MDSCompressedBoneFrame	*tempmdscompframe;

	/* General variables */

	long			numbonelist;
	long			boneindexlist[MAXBONE];
	//long			baseframenumber = 0; // should be vaiable
	long			totaltagsize = 0;
	long			numtag = 0;
	Point3			vec;

	/* Base Frame */

	if( (baseframenumber < 0) || (baseframenumber > mdsheader->lNumFrames))
	{
		printf("ERROR: Invalid BASEFRAME value.\n");
		return -1;
	}

	tempmdsframe = (MDSFrame *)((char *)mdsframe + baseframenumber * ( sizeof(MDSFrame) + mdsheader->lNumBones * sizeof(MDSCompressedBoneFrame) ) );
	
	/* Search and Convert */

	for( i = 0; i < mdsheader->lNumBones; i++)
	{

		if( currentmdsbone->ulFlags )
		{ // this bone seems a tag

			for( j = 0; j < 64; j++ ) currentmdmtag->TagName[j] = currentmdsbone->cName[j];
		
			currentmdmtag->AttachBoneNumber = mdsbonetomdm[currentmdsbone->lParentIndex];
		
			/* Build Bone Tree */

			tempmdsbone = currentmdsbone;

			numbonelist = 0;

			while( 1 ){

				boneindexlist[numbonelist++] = mdsbonetomdm[tempmdsbone->lParentIndex];

				tempmdsbone = mdsbone + tempmdsbone->lParentIndex;

				if( tempmdsbone->lParentIndex == -1) break;

			}

			currentmdmtag->NumBoneList = numbonelist;

			currentmdmtag->OffsetBoneList = sizeof(MDMTag);

			currentmdmtag->OffsetNextTag = sizeof(MDMTag) + numbonelist * sizeof(long);

			for( j = 0; j < numbonelist; j++ )
			{
				*((long *)((char *)currentmdmtag + currentmdmtag->OffsetBoneList) + j ) = boneindexlist[ numbonelist - j - 1 ];

			}

			/* Calcurate */

			tempmdscompframe = (MDSCompressedBoneFrame *)((char *)tempmdsframe + sizeof(MDSFrame) + currentmdsbone->lParentIndex * sizeof(MDSCompressedBoneFrame) );


			/* Tag's origin position */

			vec.x = currentmdsbone->fParentDist;
			vec.y = 0;
			vec.z = 0;

			pitch( &vec , tempmdscompframe->sOffsetYaw );
			yaw( &vec , tempmdscompframe->sOffsetPitch );

			currentmdmtag->OriginPosition = vec;

			/* Tag's normalized coordinates */

			vec.x = 1;
			vec.y = 0;
			vec.z = 0;

			roll( &vec , tempmdscompframe->sRoll );
			yaw( &vec , tempmdscompframe->sYaw );
			pitch( &vec , tempmdscompframe->sPitch );


			currentmdmtag->RotationMatrix[0] = vec;

			vec.x = 0;
			vec.y = 1;
			vec.z = 0;

			roll( &vec , tempmdscompframe->sRoll );
			yaw( &vec , tempmdscompframe->sYaw );
			pitch( &vec , tempmdscompframe->sPitch );

			currentmdmtag->RotationMatrix[1] = vec;

			vec.x = 0;
			vec.y = 0;
			vec.z = 1;

			roll( &vec , tempmdscompframe->sRoll );
			yaw( &vec , tempmdscompframe->sYaw );
			pitch( &vec , tempmdscompframe->sPitch );

			currentmdmtag->RotationMatrix[2] = vec;

			/* Next */

			totaltagsize += currentmdmtag->OffsetNextTag;
			numtag++;

			currentmdmtag = (MDMTag *)((char *)currentmdmtag + currentmdmtag->OffsetNextTag );
			
		} 

		currentmdsbone++;

	} 

	printf("Done.\n");

	/* Build MDM Header */

	printf("Header...");

	char		*mdmident	=	"MDMW";
	i = 0;
	while( *mdmident ) mdmheader->MDMIdent[i++] = *mdmident++;

	mdmheader->Version = 3;

	char		*mdmname	=	"models/players/temperate/allied/soldier/body.mdm";
	i = 0;
	while( *mdmname ) mdmheader->MDMName[i++] = *mdmname++;
	while(  i < 64  ) mdmheader->MDMName[i++] = '\0';

	mdmheader->LoDScale = mdsheader->fLodScale;

	mdmheader->LoDBias = mdsheader->fLodBias;

	mdmheader->NumSurface = mdsheader->lNumSurfaces;

	mdmheader->OffsetSurface = sizeof(MDMHeader);

	mdmheader->NumTag = numtag;

	mdmheader->OffsetTag = mdmheader->OffsetSurface + totalsurfacesize;

	mdmheader->OffsetEOF = mdmheader->OffsetTag + totaltagsize;

	printf("Done.\n");

	/* Successful */

	printf("MDM Convertion has been done.\n");

	printf("MDM size			:%d\n", mdmheader->OffsetEOF );

	return mdmheader->OffsetEOF;

}

/*
########################################################

Parse_MDSBone

Checks MDS Bone and removes Tags then Builds MDX Bone structure and its relation.
Returns number of MDX Bones.

########################################################
*/

char Parse_MDSBone( MDSBone *mdsbone , int NumBone , long *mdmbonetomds , long *mdsbonetomdm )
{

	/* Init general variables */

	int		i;
	char		mdxbonecount = 0;

	/* 
	Bone

	MDS has a tag as a bone while MDX does not have tag info as bone.
	So we need to get rid of tags from bones and determine which MDS 
	bone does MDM bone equal to? 
	*/

	printf("Parsing MDS Bone...");

	/* Which MDS bone is a tag? */

	MDSBone *currentmdsbone = mdsbone;

	for( i = 0; i < NumBone; i++ )
	{

	/* Assign number */

		*(mdsbonetomdm + i) = mdxbonecount;
		*(mdmbonetomds + mdxbonecount) = i;

		// if( currentmdsbone->ulFlags != 0 ) printf("MDS Bone #%d Parent Distance	:%f\n",i,currentmdsbone->fParentDist);

		if( currentmdsbone->ulFlags == 0 )
		{ // oh, this is a bone!
			mdxbonecount++;
		}

		currentmdsbone++; // Next!

	}
	/* for debug
	for( i=0; i<mdxbonecount; i++ ) printf("MDX(%d) -> MDS(%d)\n",i,*(mdmbonetomds + i));
	for( i=0; i<mdsheader->lNumBones; i++ ) printf("MDS(%d) -> MDX(%d)\n",i,*(mdsbonetomdm + i));
	*/

	printf("Done.\n");

	printf("MDS Bones			:%d (Bone %d + Tag %d)\n", NumBone ,mdxbonecount,NumBone - mdxbonecount );

	return mdxbonecount;

}

/*
########################################################

Vector Rotation Functions

yaw   : rotation with z-axis fixed
pitch : rotation with y-axis fixed
roll  : rotation with x-axis fixed

########################################################
*/

void yaw( Point3 *in, short int angle)
{

	double rad = ((double)angle) / 65536.0 * 2.0 * 3.1415;
	Point3 vec;

	vec.x = in->x * cos(rad) - in->y * sin(rad);
	vec.y = in->x * sin(rad) + in->y * cos(rad);
	vec.z = in->z;

	*in = vec;

	return;

}

void pitch( Point3 *in, short int angle)
{

	double rad = ((double)angle) / 65536.0 * 2.0 * 3.1415;
	Point3 vec;

	vec.x = in->x * cos(rad) + in->z * sin(rad);
	vec.y = in->y;
	vec.z = -in->x * sin(rad) + in->z * cos(rad);

	*in = vec;

	return;

}

void roll( Point3 *in, short int angle)
{

	double rad = ((double)angle) / 65536.0 * 2.0 * 3.1415;
	Point3 vec;

	vec.x = in->x;
	vec.y = in->y * cos(rad) - in->z * sin(rad);
	vec.z = in->y * sin(rad) + in->z * cos(rad);
	
	*in = vec;

	return;

}


/*
########################################################

Load_MDS

Load MDS into memory and return its size

########################################################
*/

int Load_MDS( char *mdsfilename , char *mdsbuf )
{

	printf("\n--- Loading MDS ---\n\n");

	int		fi;

	/* Read MDS */

	if( ( fi = open( mdsfilename , O_RDONLY ) ) == -1 ){

		printf("ERROR:Failed to open file[ %s ].\n", mdsfilename );
		return -1;

	}

	if( read( fi, mdsbuf, MAXMDSSIZE ) == -1 ){

		printf("ERROR:Failed to read file.\n");
		return -1;

	}

	close( fi );

	/* Check whether this is MDS or not */

	printf("Checking MDS Header...");

	MDSHeader *mdsbufheader = (MDSHeader *)mdsbuf;

	if( mdsbufheader->lId != 'WSDM' ){
		printf("ERROR:This MDS seems to be wrong file.\n");
		return -1;
	}

	if( mdsbufheader->lVersion != MDSVERSION ){
		printf("ERROR:This MDS version is not supported.\n");
		return -1;
	}

	if( mdsbufheader->lOffsetEnd > MAXMDSSIZE ){
		printf("ERROR:This MDS is too large.\n");
		return -1;
	}

	if( mdsbufheader->lNumFrames < 1 ){
		printf("ERROR:This MDS has no animation data.\n");
		return -1;
	}

	printf("This MDS seems to be valid.\n");

	printf("MDS has been loaded.\n");

	/* Successful */

	return mdsbufheader->lOffsetEnd;

}

/*
########################################################

Load_MDM

Load MDM into memory and return its size

########################################################
*/

int Load_MDM( char *mdmfilename , char *mdmbuf )
{

	printf("\n--- Loading MDM ---\n\n");

	int		fi;

	/* Read MDM */

	if( ( fi = open( mdmfilename , O_RDONLY ) ) == -1 ){

		printf("ERROR:Failed to open file[ %s ].\n", mdmfilename );
		return -1;

	}

	if( read( fi, mdmbuf, MAXMDMSIZE ) == -1 ){

		printf("ERROR:Failed to read file.\n");
		return -1;

	}

	close( fi );

	/* Check whether this is MDM or not */

	printf("Checking MDM Header...");

	MDMHeader *mdmbufheader = (MDMHeader *)mdmbuf;

	if( *((long *)mdmbufheader->MDMIdent) != 'WMDM' ){
		printf("ERROR:This seems to be wrong MDM file.\n");
		return -1;
	}

	if( mdmbufheader->Version != MDMVERSION ){
		printf("ERROR:This MDM version is not supported.\n");
		return -1;
	}

	if( mdmbufheader->OffsetEOF > MAXMDMSIZE ){
		printf("ERROR:This MDM is too large.\n");
		return -1;
	}

	printf("This MDM seems to be valid.\n");

	printf("MDM has been loaded.\n");

	/* Successful */

	return mdmbufheader->OffsetEOF;

}

/*
########################################################

Load_MDX

Load MDX into memory and return its size

########################################################
*/

int Load_MDX( char *mdxfilename , char *mdxbuf )
{

	printf("\n--- Loading MDX ---\n\n");

	int		fi;

	/* Read MDX */

	if( ( fi = open( mdxfilename , O_RDONLY ) ) == -1 ){

		printf("ERROR:Failed to open file[ %s ].\n", mdxfilename );
		return -1;

	}

	if( read( fi, mdxbuf, MAXMDXSIZE ) == -1 ){

		printf("ERROR:Failed to read file.\n");
		return -1;

	}

	close( fi );

	/* Check whether this is MDX or not */

	printf("Checking MDX Header...");

	MDXHeader *mdxbufheader = (MDXHeader *)mdxbuf;

	if( *((long *)mdxbufheader->MDXIdent) != 'WXDM' ){
		printf("ERROR:This seems to be wrong MDX file.\n");
		return -1;
	}

	if( mdxbufheader->MDXVersion != MDXVERSION ){
		printf("ERROR:This MDX version is not supported.\n");
		return -1;
	}

	if( mdxbufheader->OffsetEOF > MAXMDXSIZE ){
		printf("ERROR:This MDX is too large.\n");
		return -1;
	}

	printf("This MDX seems to be valid.\n");

	printf("MDX has been loaded.\n");

	/* Successful */

	return mdxbufheader->OffsetEOF;

}

/*
########################################################

Save_File

Save data and return 1 or -1 if error

########################################################
*/

int Save_File( char *data , char *filename , int size )
{

	printf("\n--- Exporting File ---\n\n");

	int fi;

	printf("Saving Data ( %s )...",filename);

	if( ( fi = open( filename , O_CREAT | O_WRONLY | O_BINARY , S_IWRITE ) ) == -1 ){

		printf("Failed to create file[ %s ].\n", filename );
		return 0;

	}

	if( write( fi, data, size ) == -1 ){

		printf("Failed to save file.\n");
		return 0;

	}

	close( fi );

	printf("Done.\n");
	
	return 1;

}

/*
########################################################

Report_MDM

Output requested MDM detail data

########################################################
*/

int Report_MDM( char *mdmbuf )
{

	printf("\nReport MDM\n\n");

	printf("***--- Header ---***\n\n");

	MDMHeader *mdm = (MDMHeader *)mdmbuf;

	printf("MDM Identification			:%s\n",mdm->MDMIdent);
	printf("MDM Version					:%d\n",mdm->Version);
	printf("MDM Name					:%s\n",mdm->MDMName);
	printf("LoD Scale					:%f\n",mdm->LoDScale);
	printf("LoD Bias					:%f\n",mdm->LoDBias);
	printf("Number of Surfaces			:%d\n",mdm->NumSurface);
	printf("Offset to Surfaces			:%d\n",mdm->OffsetSurface);
	printf("Number of Tags				:%d\n",mdm->NumTag);
	printf("Offset to Tags				:%d\n",mdm->OffsetTag);
	printf("Offset to EOF				:%d\n",mdm->OffsetEOF);

	printf("\n***--- Surface ---***\n\n");

	int sn;

	MDMSurfaceHeader *surface =  (MDMSurfaceHeader *) ((char *)mdmbuf + mdm->OffsetSurface );

	for( sn = 0; sn < mdm->NumSurface; sn++){

		printf("\n--- Surface #%0d---\n\n",sn);

		printf("Surface Identification		:%d\n",surface->SurfaceIdent);
		printf("Surface Name				:%s\n",surface->SurfaceName);
		printf("Shader Name					:%s\n",surface->ShaderName);
		printf("Shader Index				:%d\n",surface->ShaderIndex);
		printf("LoD	Minimum					:%d\n",surface->LoDMin);
		printf("Offset to Surface Header	:%d\n",surface->OffsetSurfaceHeader);
		printf("Number of Vertices			:%d\n",surface->NumVertex);
		printf("Offset to Vertices			:%d\n",surface->OffsetVertex);
		printf("Number of Triangles			:%d\n",surface->NumTriangle);
		printf("Offset to Triangles			:%d\n",surface->OffsetTriangle);
		printf("Offset to Collapsemap		:%d\n",surface->OffsetCollapsemap);
		printf("Number of Bone Refferences	:%d\n",surface->NumBoneref);
		printf("Offset to Bobe Refferences	:%d\n",surface->OffsetBoneref);
		printf("Offset to NextSurface		:%d\n",surface->OffsetNextSurface);

		surface = (MDMSurfaceHeader *) ((char *)surface + surface->OffsetNextSurface);

	}

	int tn,bn;

	MDMTag *tag = (MDMTag *)((char *)mdmbuf + mdm->OffsetTag );

	for( tn = 0; tn < mdm->NumTag; tn++)
	{
		printf("\n--- Tag #%0d ---\n\n",tn);

		printf("Name					:%s\n",tag->TagName);
		printf("Coord1					:(%f,%f,%f)\n",tag->RotationMatrix[0].x,tag->RotationMatrix[0].y,tag->RotationMatrix[0].z);
		printf("Coord2					:(%f,%f,%f)\n",tag->RotationMatrix[1].x,tag->RotationMatrix[1].y,tag->RotationMatrix[1].z);
		printf("Coord3					:(%f,%f,%f)\n",tag->RotationMatrix[2].x,tag->RotationMatrix[2].y,tag->RotationMatrix[2].z);

		printf("Attach Bone Number		:%d\n",tag->AttachBoneNumber);

		printf("Attach Bone Number		:(%f,%f,%f)\n",tag->OriginPosition.x,tag->OriginPosition.y,tag->OriginPosition.z);

		printf("Number of Bone List		:%d\n",tag->NumBoneList);
		printf("Offset to Bone List		:%d\n",tag->OffsetBoneList);
		printf("Offset to Next Tag		:%d\n",tag->OffsetNextTag);

		printf("Bone List				: ");

		for( bn = 0; bn < tag->NumBoneList; bn++){

			printf("%0d ", *((long *)((char *)tag + tag->OffsetBoneList + 4 * bn)) );

		}

		printf("\n");

		tag = (MDMTag *)((char *)tag + tag->OffsetNextTag);

	}

	return 1;

}

/*
Replace_Tag

Replaces tag's position and rotation matrix to reference mdm
Returns 1 or -1 if error
*/

int Replace_Tag( char *mdm , char *refmdm )
{

	printf("\nMDM tag replacement...");

	/* set address to pointers */

	MDMHeader	*mdmheader		=	(MDMHeader *)mdm;
	MDMTag		*mdmtag			=	(MDMTag *)((char *)mdm + mdmheader->OffsetTag);
	
	MDMHeader	*refmdmheader	=	(MDMHeader *)refmdm;
	MDMTag		*refmdmtag		=	(MDMTag *)((char *)refmdm + refmdmheader->OffsetTag);

	/* genereal valiables */

	int		i,j,replaced = 0;

	/* is number of tags same? */

	if( mdmheader->NumTag != refmdmheader->NumTag )
	{
		printf("ERROR: Number of MDM tag is not same as one of Reference MDM.\n");
		return -1;
	}

	/* Replacement */

	MDMTag		*currentmdmtag		=	mdmtag;
	MDMTag		*currentrefmdmtag	=	refmdmtag;

	for( i = 0; i < mdmheader->NumTag; i++ )
	{

		currentrefmdmtag = refmdmtag;

		for( j = 0; j < refmdmheader->NumTag; j++ )
		{

			if( strequ( currentmdmtag->TagName , currentrefmdmtag->TagName ) )
			{

				currentmdmtag->OriginPosition	=	currentrefmdmtag->OriginPosition;

				currentmdmtag->RotationMatrix[0]	= currentrefmdmtag->RotationMatrix[0];
				currentmdmtag->RotationMatrix[1]	= currentrefmdmtag->RotationMatrix[1];
				currentmdmtag->RotationMatrix[2]	= currentrefmdmtag->RotationMatrix[2];

				replaced++;

				break;

			}

			currentrefmdmtag = (MDMTag *)((char *)currentrefmdmtag + currentrefmdmtag->OffsetNextTag);
		}

		currentmdmtag = (MDMTag *)((char *)currentmdmtag + currentmdmtag->OffsetNextTag);

	}

	/* check number of replaced tags */

	if( replaced != mdmheader->NumTag )
	{
		printf("ERROR: Number of replaced tags differs number of tags.\n");
		return -1;
	}

	/* successful */

	printf("Done.\n");

	return 1;

}

/*
Replace_Bone

Replace MDX Bone length to reference MDX Bone one
*/

int	Replace_Bone( char *mdx , char *refmdx )
{

	int i,j,replaced = 0;

	MDXHeader	*mdxheader			=	(MDXHeader *)mdx;
	MDXBone		*mdxbone			=	(MDXBone *)((char *)mdx + mdxheader->OffsetBone);

	MDXHeader	*refmdxheader		=	(MDXHeader *)refmdx;
	MDXBone		*refmdxbone			=	(MDXBone *)((char *)refmdx + refmdxheader->OffsetBone);

	MDXBone		*currentmdxbone		=	mdxbone;
	MDXBone		*currentrefmdxbone	=	refmdxbone;

	if( mdxheader->NumBone != refmdxheader->NumBone )
	{
		printf("ERROR: Number of bones not same.\n");
		return -1;
	}

	for( i =0; i < refmdxheader->NumBone; i++ )
	{

		currentrefmdxbone = (MDXBone *)((char *)refmdxbone + i * sizeof(MDXBone));

		for( j = 0; j < mdxheader->NumBone; j++)
		{

			currentmdxbone = (MDXBone *)((char *)mdxbone + j * sizeof(MDXBone));

			if( strequ( currentmdxbone->BoneName , currentrefmdxbone->BoneName ) )
			{

				currentmdxbone->ParentDist = currentrefmdxbone->ParentDist;
				currentmdxbone->TorsoWeight = currentrefmdxbone->TorsoWeight;

				replaced++;

				printf("%s\n",currentmdxbone->BoneName);

				break;

			}

		}

	}

	if( replaced != refmdxheader->NumBone )
	{

		printf("ERROR: Insufficient number of replaced bones.\n");
		return -1;

	}

	return 1;

}

/*
Rescale_MDM

Rescales MDM vertex positions and tag origin positions
Returns 1 or -1 if error
*/

int Rescale_MDM( char *mdm , int rescalefactor )
{

	int i,j,k,l=0;

	float s = (float)rescalefactor / 100.0;

	MDMHeader *mdmheader = (MDMHeader *)mdm;
	MDMSurfaceHeader *mdmsurface = (MDMSurfaceHeader *)((char *)mdm + mdmheader->OffsetSurface);
	MDMTag *mdmtag = (MDMTag *)((char *)mdm + mdmheader->OffsetTag);

	MDMSurfaceHeader *currentmdmsurface = mdmsurface;
	MDMVertex *currentmdmvertex;
	MDMWeight *currentmdmweight;

	printf("%d\n",mdmheader->NumSurface);

	for( i = 0; i < mdmheader->NumSurface; i++)
	{

		currentmdmvertex = (MDMVertex *)((char *)currentmdmsurface + currentmdmsurface->OffsetVertex);

		printf("%d:%d:%d\n",i,l,currentmdmsurface->NumVertex);

		for( j = 0; j < currentmdmsurface->NumVertex; j++)
		{

			currentmdmweight = (MDMWeight *)((char *)currentmdmvertex + sizeof(MDMVertex));

			for( k = 0; k < currentmdmvertex->NumBoneWeight; k++ )
			{

				currentmdmweight->Coord.x *= s;
				currentmdmweight->Coord.y *= s;
				currentmdmweight->Coord.z *= s;

				currentmdmweight = (MDMWeight *)((char *)currentmdmweight + sizeof(MDMWeight));

				l++;

			}

			currentmdmvertex = (MDMVertex *)currentmdmweight;

		}

		currentmdmsurface = (MDMSurfaceHeader *)((char *)currentmdmsurface + currentmdmsurface->OffsetNextSurface);

	}

	MDMTag *currentmdmtag = mdmtag;

	for( i = 0; i < mdmheader->NumTag; i++ )
	{

		currentmdmtag->OriginPosition.x *= s;
		currentmdmtag->OriginPosition.y *= s;
		currentmdmtag->OriginPosition.z *= s;

		currentmdmtag = (MDMTag *)((char *)currentmdmtag + currentmdmtag->OffsetNextTag);

	}

	return 1;

}

/*
Rescale_MDX

Rescales MDX frame properties and bone parent distances
Returns 1 or -1 if error
*/

int Rescale_MDX( char *mdx , int rescalefactor )
{
	
	int i;

	float s = (float)rescalefactor / 100.0;

	MDXHeader *mdxheader = (MDXHeader *)mdx;
	MDXFrame *mdxframe = (MDXFrame *)((char *)mdx + mdxheader->OffsetFrame);
	MDXBone * mdxbone = (MDXBone *)((char *)mdx + mdxheader->OffsetBone);

	MDXFrame *currentmdxframe = mdxframe;

	for( i = 0; i < mdxheader->NumFrame; i++ )
	{

		currentmdxframe->BBoxMin.x *= s;
		currentmdxframe->BBoxMin.y *= s;
		currentmdxframe->BBoxMin.z *= s;

		currentmdxframe->BBoxMax.x *= s;
		currentmdxframe->BBoxMax.y *= s;
		currentmdxframe->BBoxMax.z *= s;

		currentmdxframe->Radius *= s;

		currentmdxframe->LocalOrigin.x *= s;
		currentmdxframe->LocalOrigin.y *= s;
		currentmdxframe->LocalOrigin.z *= s;

		currentmdxframe->OffsetParent.x *= s;
		currentmdxframe->OffsetParent.y *= s;
		currentmdxframe->OffsetParent.z *= s;

		currentmdxframe = (MDXFrame *)((char *)currentmdxframe + sizeof(MDXFrame) + mdxheader->NumBone * sizeof(MDXCompFrame));
        
	}
    
	MDXBone *currentmdxbone = mdxbone;

	for( i = 0; i < mdxheader->NumBone; i++ )
	{

		currentmdxbone->ParentDist *= s;

		currentmdxbone = (MDXBone *)((char *)currentmdxbone + sizeof(MDXBone));

	}
    
	return 1;

}

/*
Adjust_MDX_LOCALORIGINZ

Rescales MDX frame properties and bone parent distances
Returns 1 or -1 if error
*/

int Rescale_MDX_LOCALOROGINZ( char *mdx , int adjustment )
{
	
	int i;

	MDXHeader *mdxheader = (MDXHeader *)mdx;
	MDXFrame *mdxframe = (MDXFrame *)((char *)mdx + mdxheader->OffsetFrame);

	MDXFrame *currentmdxframe = mdxframe;

	for( i = 0; i < mdxheader->NumFrame; i++ )
	{

		currentmdxframe->BBoxMin.z += adjustment;

		currentmdxframe->BBoxMax.z *= adjustment;

		currentmdxframe->OffsetParent.z += adjustment;

		currentmdxframe = (MDXFrame *)((char *)currentmdxframe + sizeof(MDXFrame) + mdxheader->NumBone * sizeof(MDXCompFrame));
        
	}
  
	return 1;

}

/*
Replace_Tag

Replaces tag's position and rotation matrix to reference mdm
Returns 1 or -1 if error
*/

int Replace_Tag_All( char *mdm , char *refmdm )
{

	printf("\nMDM tag total replacement...");

	/* set address to pointers */

	MDMHeader	*mdmheader		=	(MDMHeader *)mdm;
	MDMTag		*mdmtag			=	(MDMTag *)((char *)mdm + mdmheader->OffsetTag);
	
	MDMHeader	*refmdmheader	=	(MDMHeader *)refmdm;
	MDMTag		*refmdmtag		=	(MDMTag *)((char *)refmdm + refmdmheader->OffsetTag);

	/* genereal valiables */

	int		i,j,replaced = 0;

	/* Replacement */

	char		*currentmdmtag		=	(char *)mdmtag;
	char		*currentrefmdmtag	=	(char *)refmdmtag;

	for( i = 0; i < (refmdmheader->OffsetEOF - refmdmheader->OffsetTag); i++)
	{
		*(currentmdmtag + i) = *(currentrefmdmtag + i);

	}

	/* change number and offsets */

	mdmheader->NumTag	=	refmdmheader->NumTag;
	mdmheader->OffsetEOF	=	mdmheader->OffsetTag + (refmdmheader->OffsetEOF - refmdmheader->OffsetTag);
    
	/* successful */

	printf("Done.\n");

	return 1;

}

// EOF